"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var function_1 = require("fp-ts/lib/function");
var NonEmptyArray_1 = require("fp-ts/lib/NonEmptyArray");
var OptionFn = require("fp-ts/lib/Option");
var ValidationFn = require("fp-ts/lib/Validation");
var index_1 = require("./index");
describe('AndThen Decoder', function () {
    var Direction;
    (function (Direction) {
        Direction[Direction["Up"] = 0] = "Up";
        Direction[Direction["Down"] = 1] = "Down";
        Direction[Direction["Left"] = 2] = "Left";
        Direction[Direction["Right"] = 3] = "Right";
    })(Direction || (Direction = {}));
    function directionDecoder(type) {
        switch (type) {
            case 'Up':
                return index_1.default.succeed(Direction.Up);
            case 'Down':
                return index_1.default.succeed(Direction.Down);
            case 'Left':
                return index_1.default.succeed(Direction.Left);
            case 'Right':
                return index_1.default.succeed(Direction.Right);
            default:
                return index_1.default.fail("Value \"" + type + "\" is not a valid Direction: \"Up\", \"Down\", \"Left\", \"Right\"");
        }
    }
    it('should succeed if given valid values and decoder', function () {
        var directionT = index_1.default.andThen(directionDecoder, index_1.default.string()).run('Up');
        expect(directionT).toEqual(ValidationFn.success(Direction.Up));
    });
    describe('should fail', function () {
        it('if the original decoder fails', function () {
            var directionT = index_1.default.andThen(directionDecoder, index_1.default.string()).run(true);
            var failMsg = NonEmptyArray_1.nonEmptyArray.of('Value must be a string, found "boolean" instead');
            expect(directionT).toEqual(ValidationFn.failure(failMsg));
        });
        it('if the decoder chooser fails', function () {
            var directionT = index_1.default.andThen(directionDecoder, index_1.default.string()).run('Forward');
            var failMsg = NonEmptyArray_1.nonEmptyArray.of('Value "Forward" is not a valid Direction: "Up", "Down", "Left", "Right"');
            expect(directionT).toEqual(ValidationFn.failure(failMsg));
        });
    });
});
describe('AndMap Decoder', function () {
    function user(id, username, email, isAdmin) {
        return { id: id, username: username, email: email, isAdmin: isAdmin };
    }
    var userDecoder = function_1.pipe(index_1.default.andMap(index_1.default.field('id', index_1.default.number())), index_1.default.andMap(index_1.default.field('username', index_1.default.string())), index_1.default.andMap(index_1.default.field('email', index_1.default.string())), index_1.default.andMap(index_1.default.field('isAdmin', index_1.default.boolean())))(index_1.default.succeed(function_1.curry(user)));
    it('should succeed if all decoders succeeded', function () {
        var userT = userDecoder.run({ id: 0, username: 'username', email: 'email', isAdmin: true });
        expect(userT).toEqual(ValidationFn.success({ id: 0, username: 'username', email: 'email', isAdmin: true }));
    });
    it('should provide four failures messages if all decoders failed', function () {
        var personT = userDecoder.run({});
        var failMsg = new NonEmptyArray_1.NonEmptyArray('Object missing value at key "isAdmin"', [
            'Object missing value at key "email"',
            'Object missing value at key "username"',
            'Object missing value at key "id"',
        ]);
        expect(personT).toEqual(ValidationFn.failure(failMsg));
    });
});
describe('Array Decoder', function () {
    describe('should succeed', function () {
        it('if given a boolean decoder and an array of booleans', function () {
            var booleans = index_1.default.array(index_1.default.boolean()).run([true]);
            expect(booleans).toEqual(ValidationFn.success([true]));
        });
        it('if given a number decoder and an array of numbers', function () {
            var numbers = index_1.default.array(index_1.default.number()).run([10]);
            expect(numbers).toEqual(ValidationFn.success([10]));
        });
        it('if given a string decoder and an array of strings', function () {
            var strings = index_1.default.array(index_1.default.string()).run(['string']);
            expect(strings).toEqual(ValidationFn.success(['string']));
        });
    });
    describe('should fail', function () {
        it('if given an undefined value', function () {
            var undefinedT = index_1.default.array(index_1.default.boolean()).run(undefined);
            var failMsg = NonEmptyArray_1.nonEmptyArray.of('Value must be an array, found "undefined" instead');
            expect(undefinedT).toEqual(ValidationFn.failure(failMsg));
        });
        it('if given a boolean', function () {
            var booleanT = index_1.default.array(index_1.default.boolean()).run(true);
            var failMsg = NonEmptyArray_1.nonEmptyArray.of('Value must be an array, found "boolean" instead');
            expect(booleanT).toEqual(ValidationFn.failure(failMsg));
        });
        it('if given a number', function () {
            var numberT = index_1.default.array(index_1.default.number()).run(10);
            var failMsg = NonEmptyArray_1.nonEmptyArray.of('Value must be an array, found "number" instead');
            expect(numberT).toEqual(ValidationFn.failure(failMsg));
        });
        it('if given a string', function () {
            var stringT = index_1.default.array(index_1.default.string()).run('string');
            var failMsg = NonEmptyArray_1.nonEmptyArray.of('Value must be an array, found "string" instead');
            expect(stringT).toEqual(ValidationFn.failure(failMsg));
        });
        it('if given an object', function () {
            var objectT = index_1.default.array(index_1.default.boolean()).run({});
            var failMsg = NonEmptyArray_1.nonEmptyArray.of('Value must be an array, found "object" instead');
            expect(objectT).toEqual(ValidationFn.failure(failMsg));
        });
        it('if given a string decoder and an array of booleans', function () {
            var booleans = index_1.default.array(index_1.default.string()).run([true]);
            var failMsg = NonEmptyArray_1.nonEmptyArray.of('Value must be a string, found "boolean" instead at index #0');
            expect(booleans).toEqual(ValidationFn.failure(failMsg));
        });
        it('if given a string decoder and an array of mixed values', function () {
            var booleans = index_1.default.array(index_1.default.string()).run([true, 10]);
            var firstFailMsg = NonEmptyArray_1.nonEmptyArray.of('Value must be a string, found "boolean" instead at index #0');
            var secondFailMsg = NonEmptyArray_1.nonEmptyArray.of('Value must be a string, found "number" instead at index #1');
            var failMsgs = firstFailMsg.concat(secondFailMsg);
            expect(booleans).toEqual(ValidationFn.failure(failMsgs));
        });
    });
});
describe('At Decoder', function () {
    it('should succeed if object has nested keys', function () {
        var booleanT = index_1.default.at(['first', 'boolean'], index_1.default.boolean()).run({ first: { boolean: true } });
        expect(booleanT).toEqual(ValidationFn.success(true));
    });
});
describe('Boolean Decoder', function () {
    it('should succeed if given a boolean', function () {
        var booleanT = index_1.default.boolean().run(true);
        expect(booleanT).toEqual(ValidationFn.success(true));
    });
    it('should fail if given an undefined value', function () {
        var undefinedT = index_1.default.boolean().run(undefined);
        var failMsg = NonEmptyArray_1.nonEmptyArray.of('Value must be a boolean, found "undefined" instead');
        expect(undefinedT).toEqual(ValidationFn.failure(failMsg));
    });
    it('should fail if given a number', function () {
        var numberT = index_1.default.boolean().run(10);
        var failMsg = NonEmptyArray_1.nonEmptyArray.of('Value must be a boolean, found "number" instead');
        expect(numberT).toEqual(ValidationFn.failure(failMsg));
    });
    it('should fail if given a string', function () {
        var stringT = index_1.default.boolean().run('string');
        var failMsg = NonEmptyArray_1.nonEmptyArray.of('Value must be a boolean, found "string" instead');
        expect(stringT).toEqual(ValidationFn.failure(failMsg));
    });
    it('should fail if given an object', function () {
        var objectT = index_1.default.boolean().run({});
        var failMsg = NonEmptyArray_1.nonEmptyArray.of('Value must be a boolean, found "object" instead');
        expect(objectT).toEqual(ValidationFn.failure(failMsg));
    });
});
describe('Fail Decoder', function () {
    it('should fail no matter what value is given', function () {
        var booleanT = index_1.default.fail('Custom fail message').run([true]);
        var failMsg = NonEmptyArray_1.nonEmptyArray.of('Custom fail message');
        expect(booleanT).toEqual(ValidationFn.failure(failMsg));
    });
});
describe('Field Decoder', function () {
    it('should succeed if given an object with a boolean', function () {
        var booleanT = index_1.default.field('boolean', index_1.default.boolean()).run({ boolean: true });
        expect(booleanT).toEqual(ValidationFn.success(true));
    });
    describe('should fail', function () {
        it('if given an object with another value', function () {
            var stringT = index_1.default.field('boolean', index_1.default.boolean()).run({ boolean: 'true' });
            var failMsg = NonEmptyArray_1.nonEmptyArray.of('Value must be a boolean, found "string" instead at key "boolean"');
            expect(stringT).toEqual(ValidationFn.failure(failMsg));
        });
        it('if given an object with missing key', function () {
            var emptyT = index_1.default.field('boolean', index_1.default.boolean()).run({});
            var failMsg = NonEmptyArray_1.nonEmptyArray.of('Object missing value at key "boolean"');
            expect(emptyT).toEqual(ValidationFn.failure(failMsg));
        });
        it('if given a boolean', function () {
            var booleanT = index_1.default.field('boolean', index_1.default.boolean()).run(true);
            var failMsg = NonEmptyArray_1.nonEmptyArray.of('Value must be an object, found "boolean" instead');
            expect(booleanT).toEqual(ValidationFn.failure(failMsg));
        });
    });
});
describe('Index Decoder', function () {
    it('should succeed if given an array with a boolean', function () {
        var booleanT = index_1.default.index(0, index_1.default.boolean()).run([true]);
        expect(booleanT).toEqual(ValidationFn.success(true));
    });
    describe('should fail', function () {
        it('if given an array with another value at that index', function () {
            var stringT = index_1.default.index(0, index_1.default.boolean()).run(['true']);
            var failMsg = NonEmptyArray_1.nonEmptyArray.of('Value must be a boolean, found "string" instead at index #0');
            expect(stringT).toEqual(ValidationFn.failure(failMsg));
        });
        it('if given an array with missing index', function () {
            var emptyT = index_1.default.index(0, index_1.default.boolean()).run([]);
            var failMsg = NonEmptyArray_1.nonEmptyArray.of('Array missing value at index #0');
            expect(emptyT).toEqual(ValidationFn.failure(failMsg));
        });
        it('if given a boolean', function () {
            var booleanT = index_1.default.index(0, index_1.default.boolean()).run(true);
            var failMsg = NonEmptyArray_1.nonEmptyArray.of('Value must be an array, found "boolean" instead');
            expect(booleanT).toEqual(ValidationFn.failure(failMsg));
        });
    });
});
describe('Map Decoder', function () {
    it('should return the transformed decoded value', function () {
        var numberT = index_1.default.map(Math.abs, index_1.default.number()).run(-3);
        expect(numberT).toEqual(ValidationFn.success(3));
    });
});
describe('Map2 Decoder', function () {
    function point(x, y) {
        return { x: x, y: y };
    }
    var pointDecoder = index_1.default.map2(point, index_1.default.field('x', index_1.default.number()), index_1.default.field('y', index_1.default.number()));
    it('should succeed if all decoders succeeded', function () {
        var pointT = pointDecoder.run({ x: 2, y: 4 });
        expect(pointT).toEqual(ValidationFn.success({ x: 2, y: 4 }));
    });
    it('should fail if the first decoder failed', function () {
        var pointT = pointDecoder.run({ x: '2', y: 4 });
        var failMsg = NonEmptyArray_1.nonEmptyArray.of('Value must be a number, found "string" instead at key "x"');
        expect(pointT).toEqual(ValidationFn.failure(failMsg));
    });
    it('should fail if the second decoder failed', function () {
        var pointT = pointDecoder.run({ x: 2, y: '4' });
        var failMsg = NonEmptyArray_1.nonEmptyArray.of('Value must be a number, found "string" instead at key "y"');
        expect(pointT).toEqual(ValidationFn.failure(failMsg));
    });
    it('should provide two failures messages if all decoders failed', function () {
        var pointT = pointDecoder.run({ x: '2', y: '4' });
        var failMsg = new NonEmptyArray_1.NonEmptyArray('Value must be a number, found "string" instead at key "x"', ['Value must be a number, found "string" instead at key "y"']);
        expect(pointT).toEqual(ValidationFn.failure(failMsg));
    });
});
describe('Map3 Decoder', function () {
    function person(name, height, isAdult) {
        return { name: name, height: height, isAdult: isAdult };
    }
    var personDecoder = index_1.default.map3(person, index_1.default.field('name', index_1.default.string()), index_1.default.at(['info', 'height'], index_1.default.number()), index_1.default.at(['info', 'isAdult'], index_1.default.boolean()));
    it('should succeed if all decoders succeeded', function () {
        var personT = personDecoder.run({ name: 'Tester', info: { height: 1.8, isAdult: true } });
        expect(personT).toEqual(ValidationFn.success({ name: 'Tester', height: 1.8, isAdult: true }));
    });
    it('should fail if the first decoder failed', function () {
        var personT = personDecoder.run({ info: { height: 1.8, isAdult: true } });
        var failMsg = NonEmptyArray_1.nonEmptyArray.of('Object missing value at key "name"');
        expect(personT).toEqual(ValidationFn.failure(failMsg));
    });
    it('should fail if the second decoder failed', function () {
        var personT = personDecoder.run({ name: 'Tester', info: { isAdult: true } });
        var failMsg = NonEmptyArray_1.nonEmptyArray.of('Object missing value at key "height" at key "info"');
        expect(personT).toEqual(ValidationFn.failure(failMsg));
    });
    it('should fail if the third decoder failed', function () {
        var personT = personDecoder.run({ name: 'Tester', info: { height: 1.8 } });
        var failMsg = NonEmptyArray_1.nonEmptyArray.of('Object missing value at key "isAdult" at key "info"');
        expect(personT).toEqual(ValidationFn.failure(failMsg));
    });
    it('should provide two failures messages if the first and second decoders failed', function () {
        var personT = personDecoder.run({ info: { isAdult: true } });
        var failMsg = new NonEmptyArray_1.NonEmptyArray('Object missing value at key "name"', ['Object missing value at key "height" at key "info"']);
        expect(personT).toEqual(ValidationFn.failure(failMsg));
    });
    it('should provide two failures messages if the first and third decoders failed', function () {
        var personT = personDecoder.run({ info: { height: 1.8 } });
        var failMsg = new NonEmptyArray_1.NonEmptyArray('Object missing value at key "name"', ['Object missing value at key "isAdult" at key "info"']);
        expect(personT).toEqual(ValidationFn.failure(failMsg));
    });
    it('should provide two failures messages if the second and third decoders failed', function () {
        var personT = personDecoder.run({ name: 'Tester', info: {} });
        var failMsg = new NonEmptyArray_1.NonEmptyArray('Object missing value at key "height" at key "info"', ['Object missing value at key "isAdult" at key "info"']);
        expect(personT).toEqual(ValidationFn.failure(failMsg));
    });
    it('should provide three failures messages if all decoders failed', function () {
        var personT = personDecoder.run({ info: {} });
        var failMsg = new NonEmptyArray_1.NonEmptyArray('Object missing value at key "name"', [
            'Object missing value at key "height" at key "info"',
            'Object missing value at key "isAdult" at key "info"',
        ]);
        expect(personT).toEqual(ValidationFn.failure(failMsg));
    });
});
describe('NonEmptyArray', function () {
    it('should succeed if given a boolean decoder and a non-empty array of booleans', function () {
        var booleansResult = index_1.default.nonEmptyArray(index_1.default.boolean()).run([true]);
        var booleans = NonEmptyArray_1.nonEmptyArray.of(true);
        expect(booleansResult).toEqual(ValidationFn.success(booleans));
    });
    it('should fail if given an empty array', function () {
        var booleansResult = index_1.default.nonEmptyArray(index_1.default.boolean()).run([]);
        var failMsg = NonEmptyArray_1.nonEmptyArray.of('Array must have at least one value');
        expect(booleansResult).toEqual(ValidationFn.failure(failMsg));
    });
});
describe('Null Decoder', function () {
    it('should succeed if given a null value', function () {
        var booleanT = index_1.default.null(true).run(null);
        expect(booleanT).toEqual(ValidationFn.success(true));
    });
    it('should fail if given another value', function () {
        var booleanT = index_1.default.null(true).run(true);
        var failMsg = NonEmptyArray_1.nonEmptyArray.of('Value must be a null, found "boolean" instead');
        expect(booleanT).toEqual(ValidationFn.failure(failMsg));
    });
});
describe('Nullable Decoder', function () {
    describe('should succeed', function () {
        it('if given a null value', function () {
            var booleanT = index_1.default.nullable(index_1.default.boolean()).run(null);
            expect(booleanT).toEqual(ValidationFn.success(OptionFn.none));
        });
        it('if given a boolean decoder and a boolean', function () {
            var booleanT = index_1.default.nullable(index_1.default.boolean()).run(true);
            expect(booleanT).toEqual(ValidationFn.success(OptionFn.some(true)));
        });
        it('if given a number decoder and a number', function () {
            var numberT = index_1.default.nullable(index_1.default.number()).run(10);
            expect(numberT).toEqual(ValidationFn.success(OptionFn.some(10)));
        });
        it('if given a string decoder and a string', function () {
            var stringT = index_1.default.nullable(index_1.default.string()).run('string');
            expect(stringT).toEqual(ValidationFn.success(OptionFn.some('string')));
        });
    });
    describe('should fail', function () {
        it('if given an undefined value', function () {
            var undefinedT = index_1.default.nullable(index_1.default.boolean()).run(undefined);
            var failMsg = NonEmptyArray_1.nonEmptyArray.of('Value must be a boolean, found "undefined" instead');
            expect(undefinedT).toEqual(ValidationFn.failure(failMsg));
        });
        it('if given an object', function () {
            var objectT = index_1.default.nullable(index_1.default.boolean()).run({});
            var failMsg = NonEmptyArray_1.nonEmptyArray.of('Value must be a boolean, found "object" instead');
            expect(objectT).toEqual(ValidationFn.failure(failMsg));
        });
    });
});
describe('Number Decoder', function () {
    it('should succeed if given a decimal number', function () {
        var decimal = index_1.default.number().run(3.1);
        expect(decimal).toEqual(ValidationFn.success(3.1));
    });
    it('should succeed if given a hexadecimal number', function () {
        var hexadecimal = index_1.default.number().run(0xf00d);
        expect(hexadecimal).toEqual(ValidationFn.success(0xf00d));
    });
    it('should succeed if given a binary number', function () {
        var hexadecimal = index_1.default.number().run(10);
        expect(hexadecimal).toEqual(ValidationFn.success(10));
    });
    it('should succeed if given an octal number', function () {
        var octal = index_1.default.number().run(484);
        expect(octal).toEqual(ValidationFn.success(484));
    });
    it('should fail if given an undefined value', function () {
        var undefinedT = index_1.default.number().run(undefined);
        var failMsg = NonEmptyArray_1.nonEmptyArray.of('Value must be a number, found "undefined" instead');
        expect(undefinedT).toEqual(ValidationFn.failure(failMsg));
    });
    it('should fail if given a boolean', function () {
        var booleanT = index_1.default.number().run(true);
        var failMsg = NonEmptyArray_1.nonEmptyArray.of('Value must be a number, found "boolean" instead');
        expect(booleanT).toEqual(ValidationFn.failure(failMsg));
    });
    it('should fail if given a string', function () {
        var stringT = index_1.default.number().run('string');
        var failMsg = NonEmptyArray_1.nonEmptyArray.of('Value must be a number, found "string" instead');
        expect(stringT).toEqual(ValidationFn.failure(failMsg));
    });
    it('should fail if given an object', function () {
        var objectT = index_1.default.number().run({});
        var failMsg = NonEmptyArray_1.nonEmptyArray.of('Value must be a number, found "object" instead');
        expect(objectT).toEqual(ValidationFn.failure(failMsg));
    });
});
describe('OneOf Decoder', function () {
    var numberDecoder = NonEmptyArray_1.nonEmptyArray.of(index_1.default.number());
    var nullDecoder = NonEmptyArray_1.nonEmptyArray.of(index_1.default.null(0));
    var failDecoder = NonEmptyArray_1.nonEmptyArray.of(index_1.default.fail('Expecting a number or a null value'));
    var decoders = numberDecoder.concat(nullDecoder).concat(failDecoder);
    var badIntDecoder = index_1.default.oneOf(decoders);
    it('should succeed if the first decoder succeeds', function () {
        var numberT = badIntDecoder.run(2);
        expect(numberT).toEqual(ValidationFn.success(2));
    });
    it('should succeed if the first decoder fails but the second succeeds', function () {
        var numberT = badIntDecoder.run(null);
        expect(numberT).toEqual(ValidationFn.success(0));
    });
    it('should fail if the first and second decoders fails', function () {
        var booleanT = badIntDecoder.run(true);
        var failMsg = NonEmptyArray_1.nonEmptyArray.of('Expecting a number or a null value');
        expect(booleanT).toEqual(ValidationFn.failure(failMsg));
    });
});
describe('Optional Field Decoder', function () {
    describe('should succeed', function () {
        it('with some if given an object with the given field', function () {
            var booleanT = index_1.default.optionalField('boolean', index_1.default.boolean()).run({ boolean: true });
            expect(booleanT).toEqual(ValidationFn.success(OptionFn.some(true)));
        });
        it('with none if given an object with missing key', function () {
            var emptyT = index_1.default.optionalField('boolean', index_1.default.boolean()).run({});
            expect(emptyT).toEqual(ValidationFn.success(OptionFn.none));
        });
    });
    describe('should fail', function () {
        it('if given an object with another value', function () {
            var stringT = index_1.default.optionalField('boolean', index_1.default.boolean()).run({ boolean: 'true' });
            var failMsg = NonEmptyArray_1.nonEmptyArray.of('Value must be a boolean, found "string" instead at key "boolean"');
            expect(stringT).toEqual(ValidationFn.failure(failMsg));
        });
        it('if given a boolean', function () {
            var booleanT = index_1.default.optionalField('boolean', index_1.default.boolean()).run(true);
            var failMsg = NonEmptyArray_1.nonEmptyArray.of('Value must be an object, found "boolean" instead');
            expect(booleanT).toEqual(ValidationFn.failure(failMsg));
        });
    });
});
describe('String Decoder', function () {
    it('should succeed if given a string', function () {
        var stringT = index_1.default.string().run('string');
        expect(stringT).toEqual(ValidationFn.success('string'));
    });
    it('should fail if given an undefined value', function () {
        var undefinedT = index_1.default.string().run(undefined);
        var failMsg = NonEmptyArray_1.nonEmptyArray.of('Value must be a string, found "undefined" instead');
        expect(undefinedT).toEqual(ValidationFn.failure(failMsg));
    });
    it('should fail if given a boolean', function () {
        var booleanT = index_1.default.string().run(true);
        var failMsg = NonEmptyArray_1.nonEmptyArray.of('Value must be a string, found "boolean" instead');
        expect(booleanT).toEqual(ValidationFn.failure(failMsg));
    });
    it('should fail if given a number', function () {
        var numberT = index_1.default.string().run(10);
        var failMsg = NonEmptyArray_1.nonEmptyArray.of('Value must be a string, found "number" instead');
        expect(numberT).toEqual(ValidationFn.failure(failMsg));
    });
    it('should fail if given an object', function () {
        var objectT = index_1.default.string().run({});
        var failMsg = NonEmptyArray_1.nonEmptyArray.of('Value must be a string, found "object" instead');
        expect(objectT).toEqual(ValidationFn.failure(failMsg));
    });
});
describe('Succeed Decoder', function () {
    it('should succeed no matter what value is given', function () {
        var booleanT = index_1.default.succeed(0).run([true]);
        expect(booleanT).toEqual(ValidationFn.success(0));
    });
});
describe('Unknown Decoder', function () {
    it('should succeed while not transforming the value', function () {
        var booleanT = index_1.default.unknown().run(true);
        expect(booleanT).toEqual(ValidationFn.success(true));
    });
});
describe('Abstract Decoder', function () {
    test('should throw if still abstract', function () {
        var decoder = new index_1.default();
        expect(function () { return decoder.run({}); }).toThrow();
    });
});
